==> deeno-0.1.tar.gz: deeno-0.1/src/deeno/db.py <==
from functools import partial, wraps
from contextlib import contextmanager
import logging
import re
import sqlite3
import threading
from collections import namedtuple

from deeno import sql, DeenoException

try:
    import psycopg2
    from psycopg2.extras import NamedTupleCursor
except ImportError as e:
    psycopg2 = e

log = logging.getLogger(__name__)


class NoMatchingRecord(DeenoException):
    pass


class Record(dict):

    def __init__(self, table, *args, **kwargs):
        super(Record, self).__init__(*args, **kwargs)
        self.__dict__['_table'] = table
        assert set(table.pk).issubset(self)

    def __getattr__(self, name):
        try:
            return self.__getitem__(name)
        except KeyError:
            raise AttributeError(name)

    def __setattr__(self, name, value):
        if name in self.__dict__:
            self.__dict__[name] = value
        else:
            self.__setitem__(name, value)

    def __delattr__(self, name):
        if name in self.__dict__:
            self.__dict__.__delitem__(name)
        else:
            self.__delitem__(name)

    def __setitem__(self, name, value):
        if name in self._table.pk:
            raise AttributeError('primary key values are immutable')
        super(Record, self).__setitem__(name, value)

    def __delitem__(self):
        raise AttributeError('record attributes cannot be removed')

    def save(self):
        if not self._table.pk:
            raise AttributeError('only records from tables with a primary key '
                'can be saved')
        if set(self._table.pk) == set(self.keys()):
            log.debug('nothing to save')
            return
        where, values = {}, {}
        for k, v in self.items():
            if k in self._table.pk:
                where[k] = v
            else:
                values[k] = v
        self._table.update(values, where=where)

    def delete(self):
        if not self._table.pk:
            raise AttributeError('only records from tables with a primary key '
                'can be deleted')
        where = dict((k, self[k]) for k in self._table.pk)
        self._table.delete(where=where)


def table_only_operation(method):

    @wraps(method)
    def wrapper(relation, *args, **kwargs):
        if relation.type != 'table':
            raise AttributeError('invoked table-only operation %r on '
                'non-table relation %r' % (method.__name__, relation.name))
        return method(relation, *args, **kwargs)

    return wrapper


def primary_key_only_operation(method):

    @wraps(method)
    def wrapper(relation, *args, **kwargs):
        if not relation.pk:
            raise AttributeError('invoked primary-key-requiring operation %r '
                'on relation without a primary key')
        return method(relation, *args, **kwargs)

    return wrapper


class Relation(object):

    _rx_join = re.compile(r'(?:(?:cross_)|'
        '(?:(?:natural_)?(?:inner_)?(?:left_|right_|full_)?(?:outer_)?)'
        ')join$', re.I)

    def __init__(self, db, name, rtype, pk):
        self.db = db
        self.name = name if rtype == 'join' else db.quote(name)
        self.type = rtype
        self.pk = pk or ()

    def __str__(self):
        return self.name

    def get(self, **kwargs):
        try:
            row = next(self.select(where=kwargs))
        except StopIteration:
            raise NoMatchingRecord(kwargs)
        return Record(self, **row._asdict())

    def select(self, *args, **kwargs):
        q, p = sql.select(self, *args, **kwargs)
        rows = self.db.fetchall(q, p)
        return rows

    @table_only_operation
    @primary_key_only_operation
    def new(self, **kwargs):
        row = self.db.insert_and_return_row(self, kwargs)
        return Record(self, **row._asdict())

    @table_only_operation
    def delete(self, *args, **kwargs):
        q, p = sql.delete(self, *args, **kwargs)
        if kwargs.get('returning'):
            result = self.db.fetchall(q, p)
        else:
            result = self.db.execute(q, p)
        return result

    @table_only_operation
    def update(self, *args, **kwargs):
        q, p = sql.update(self, *args, **kwargs)
        if kwargs.get('returning'):
            result = self.db.fetchall(q, p)
        else:
            result = self.db.execute(q, p)
        return result

    @table_only_operation
    def insert(self, *args, **kwargs):
        q, p = sql.insert(self, *args, **kwargs)
        if kwargs.get('returning'):
            result = self.db.fetchall(q, p)
        else:
            result = self.db.execute(q, p)
        self.db._tl.last_insert_table = self
        return result

    def __getattr__(self, name):
        if self._rx_join.match(name):
            mode = name.replace('_', ' ').upper()
            return partial(self._join, mode)
        raise AttributeError(name)

    def _join(self, mode, name, on=None, using=None):
        name = self.db.quote(name)
        name = sql.join(self, mode, name, on, using)
        return Relation(self.db, name, 'join', None)


class Relatiomat(object):

    def __init__(self, db):
        self.__db = db
        self._cache = {}

    def __get_relation(self, name):
        name = '.'.join(name.split('__', 1))
        try:
            return self._cache[name]
        except KeyError:
            rtype, pk = self.__db.get_relation_info(name)
            r = Relation(self.__db, name, rtype, pk or ())
            self._cache[name] = r
            return r

    __call__ = __get_relation
    __getattr__ = __get_relation
    __getitem__ = __get_relation


def logquery(meth):

    rx_strip_query = re.compile(r' *\n *')

    @wraps(meth)
    def wrapper(db, query, *args):
        stripped = rx_strip_query.sub(' ', query.strip())
        params, = args or ((),)
        log.debug('%s: %s %r' % (meth.__name__, stripped, params))
        return meth(db, query, *args)

    return wrapper


class Database(object):

    thread_local_config = False

    placeholder = None
    quote_char = '"'
    schema_separator_char = None
    script_func = 'execute'

    def __init__(self, config=None):
        self.config = config
        self._tl = threading.local()

    def connect(self):
        raise NotImplementedError

    def disconnect(self):
        c = getattr(self._tl, 'connection', None)
        if c is not None:
            c.close()
        self._tl.connection = None

    def invalidate_relation_cache(self):
        """Invalidate the relation meta information cache.

        To be called when the database schema context or primary keys of tables
        have changed so that the cached meta information for relations may not
        be valid anymore.

        Note that this only affects the current thread's relation info cache.

        """
        self.relatiomat._cache.clear()

    @property
    def relatiomat(self):
        if getattr(self._tl, 'relatiomat', None) is None:
            self._tl.relatiomat = Relatiomat(self)
        return self._tl.relatiomat

    r = relatiomat
    t = relatiomat

    @property
    def connection(self):
        if getattr(self._tl, 'connection', None) is None:
            self.connect()
        return self._tl.connection

    def get_relation_info(self, name):
        raise NotImplementedError

    def begin(self):
        self.execute('BEGIN')

    def rollback(self):
        self.connection.rollback()

    def commit(self):
        self.connection.commit()

    @logquery
    def execute(self, query, params=()):
        c = self.connection.cursor()
        c.execute(query, params)
        return None if c.rowcount == -1 else c.rowcount

    @logquery
    def script(self, query):
        """Execute a parameterless query with multiple SQL statements

        This is only useful if a subclass' `execute` does not allow multiple
        statements in queries (e.g. `SQLiteDatabase`).

        """
        c = self.connection.cursor()
        f = getattr(c, self.script_func)
        f(query)

    @logquery
    def fetchall(self, query, params=()):
        c = self.connection.cursor()
        c.execute(query, params)
        return iter(c)

    @logquery
    def fetchone(self, query, params=()):
        c = self.connection.cursor()
        c.execute(query, params)
        return c.fetchone()

    @contextmanager
    def transaction(self):
        self.begin()
        try:
            yield
        except:
            self.rollback()
            raise
        self.commit()

    def quote(self, s):

        qc = self.quote_char
        ssc = self.schema_separator_char or ''
        parts = s.split(ssc) if ssc else [s]
        parts = ['%s%s%s' % (qc, part, qc) for part in parts]
        return ssc.join(parts)

    def insert_and_return_row(self, relation, row):
        raise NotImplementedError


class SQLiteDatabase(Database):

    placeholder = '?'
    script_func = 'executescript'

    def __init__(self, *args, **kwargs):
        super(SQLiteDatabase, self).__init__(*args, **kwargs)
        self._row_types = {}

    def _row_factory(self, cursor, row):
        try:
            row_type = self._row_types[cursor.description]
        except KeyError:
            colnames = tuple(x[0] for x in cursor.description)
            row_type = namedtuple('DeenoSqliteRow', colnames)
            self._row_types[cursor.description] = row_type
        return row_type._make(row)

    def connect(self):
        self._tl.connection = sqlite3.connect(self.config['path'])
        self._tl.connection.row_factory = self._row_factory
        self._tl.connection.isolation_level = None

    def get_relation_info(self, name):

        q = 'SELECT type FROM sqlite_master WHERE name = ?'
        p = [name]
        r = self.fetchone(q, p)
        if r is None:
            raise ValueError('no such relation: %r' % name)
        if r.type not in ('table', 'view'):
            raise ValueError('unknown relation type: %r (%r)' % (r.type, name))

        pk = None
        if r.type == 'table':
            q = 'pragma table_info(%s)' % self.quote(name)
            rows = self.fetchall(q)
            pk = tuple(row.name for row in rows if row.pk)

        return r.type, pk

    def insert_and_return_row(self, relation, row):
        q, p = sql.insert(relation, row=row)
        c = self.connection.cursor()
        c.execute(q, p)
        rowid = c.lastrowid
        rows = relation.select(where={'rowid': rowid})
        return next(rows)


class PostgresDatabase(Database):

    placeholder = '%s'
    schema_separator_char = '.'

    def __init__(self, *args, **kwargs):
        if isinstance(psycopg2, ImportError):
            raise psycopg2
        super(PostgresDatabase, self).__init__(*args, **kwargs)

    def connect(self):
        self._tl.connection = psycopg2.connect(host=self.config['host'],
            database=self.config['database'], user=self.config['user'])
        self._tl.connection.cursor_factory = NamedTupleCursor
        self._tl.connection.autocommit = True

    def begin(self):
        self.connection.autocommit = False

    def rollback(self):
        self.connection.rollback()
        self.connection.autocommit = True

    def commit(self):
        self.connection.commit()
        self.connection.autocommit = True

    def get_relation_info(self, name):

        if '.' in name:
            schema, name = name.split('.', 1)
            schemas = [schema]
        else:
            q = "SHOW search_path"
            row = self.fetchone(q)
            schemas = row[0].replace(',', ' ').split()

        log.debug('look for %r in schemas %r' % (name, schemas))

        # Since the same base relation name may occur in multiple schemas, we
        # have to walk the search path to find the correct schema for a
        # relation.

        rtype, pk = None, None
        for schema in schemas:
            p = [schema, name]
            q = """
                SELECT 1 AS x FROM pg_catalog.pg_views
                WHERE schemaname = %s AND viewname = %s
            """
            row = self.fetchone(q, p)
            if row:
                rtype = 'view'
                break
            q = """
                SELECT 1 AS x FROM pg_catalog.pg_tables
                WHERE schemaname = %s AND tablename = %s
            """
            row = self.fetchone(q, p)
            if row:
                rtype = 'table'
                break

        if rtype is None:
            raise ValueError('no view/table %r in current search path' % name)

        if rtype == 'table':
            q = """
                SELECT
                    a.attname
                FROM
                    pg_catalog.pg_attribute a,
                    pg_catalog.pg_class c,
                    pg_catalog.pg_index i
                WHERE
                    c.oid = %s::regclass AND
                    indrelid = c.oid AND
                    a.attrelid = c.oid AND
                    a.attnum = any(i.indkey)
                    AND indisprimary
            """
            p = [self.quote('%s.%s' % (schema, name))]
            rows = self.fetchall(q, p)
            pk = tuple(r[0] for r in rows)

        return rtype, pk

    def insert_and_return_row(self, relation, row):
        rows = relation.insert(row=row, returning='*')
        return next(rows)

==> deeno-0.3.tar.gz: deeno-0.3/src/deeno/db.py <==
from functools import partial, wraps
from contextlib import contextmanager
import logging
import re
import sqlite3
import threading
from collections import namedtuple

from deeno import sql, DeenoException

try:
    import psycopg2
    from psycopg2.extras import NamedTupleCursor
except ImportError as e:
    psycopg2 = e

log = logging.getLogger(__name__)


class NoMatchingRecord(DeenoException):
    pass


class RelationNotFound(DeenoException):
    pass


class Record(dict):

    def __init__(self, table, *args, **kwargs):
        super(Record, self).__init__(*args, **kwargs)
        self.__dict__['_table'] = table
        assert set(table.pk).issubset(self)

    def __getattr__(self, name):
        try:
            return self.__getitem__(name)
        except KeyError:
            raise AttributeError(name)

    def __setattr__(self, name, value):
        if name in self.__dict__:
            self.__dict__[name] = value
        else:
            self.__setitem__(name, value)

    def __delattr__(self, name):
        if name in self.__dict__:
            self.__dict__.__delitem__(name)
        else:
            self.__delitem__(name)

    def __setitem__(self, name, value):
        if name in self._table.pk:
            raise AttributeError('primary key values are immutable')
        super(Record, self).__setitem__(name, value)

    def __delitem__(self):
        raise AttributeError('record attributes cannot be removed')

    def save(self):
        if not self._table.pk:
            raise AttributeError('only records from tables with a primary key '
                'can be saved')
        if set(self._table.pk) == set(self.keys()):
            log.debug('nothing to save')
            return
        where, values = {}, {}
        for k, v in self.items():
            if k in self._table.pk:
                where[k] = v
            else:
                values[k] = v
        self._table.update(values, where=where)

    def delete(self):
        if not self._table.pk:
            raise AttributeError('only records from tables with a primary key '
                'can be deleted')
        where = dict((k, self[k]) for k in self._table.pk)
        self._table.delete(where=where)


def table_only_operation(method):

    @wraps(method)
    def wrapper(relation, *args, **kwargs):
        if relation.type != 'table':
            raise AttributeError('invoked table-only operation %r on '
                'non-table relation %r' % (method.__name__, relation.name))
        return method(relation, *args, **kwargs)

    return wrapper


def primary_key_only_operation(method):

    @wraps(method)
    def wrapper(relation, *args, **kwargs):
        if not relation.pk:
            raise AttributeError('invoked primary-key-requiring operation %r '
                'on relation without a primary key' % method.__name__)
        return method(relation, *args, **kwargs)

    return wrapper


class Relation(object):

    _rx_join = re.compile(r'(?:(?:cross_)|'
        '(?:(?:natural_)?(?:inner_)?(?:left_|right_|full_)?(?:outer_)?)'
        ')join$', re.I)

    def __init__(self, db, name, rtype, pk):
        self.db = db
        self.name = name if rtype == 'join' else db.quote(name)
        self.type = rtype
        self.pk = pk or ()

    def __str__(self):
        return self.name

    def get(self, **kwargs):
        try:
            row = next(self.select(where=kwargs))
        except StopIteration:
            raise NoMatchingRecord(kwargs)
        return Record(self, **row._asdict())

    def select(self, *args, **kwargs):
        q, p = sql.select(self, *args, **kwargs)
        rows = self.db.fetchall(q, p)
        return rows

    @table_only_operation
    @primary_key_only_operation
    def new(self, **kwargs):
        row = self.db.insert_and_return_row(self, kwargs)
        return Record(self, **row._asdict())

    @table_only_operation
    def delete(self, *args, **kwargs):
        q, p = sql.delete(self, *args, **kwargs)
        if kwargs.get('returning'):
            result = self.db.fetchall(q, p)
        else:
            result = self.db.execute(q, p)
        return result

    @table_only_operation
    def update(self, *args, **kwargs):
        q, p = sql.update(self, *args, **kwargs)
        if kwargs.get('returning'):
            result = self.db.fetchall(q, p)
        else:
            result = self.db.execute(q, p)
        return result

    @table_only_operation
    def insert(self, *args, **kwargs):
        q, p = sql.insert(self, *args, **kwargs)
        if kwargs.get('returning'):
            result = self.db.fetchall(q, p)
        else:
            result = self.db.execute(q, p)
        self.db._tl.last_insert_table = self
        return result

    def __getattr__(self, name):
        if self._rx_join.match(name):
            mode = name.replace('_', ' ').upper()
            return partial(self._join, mode)
        raise AttributeError(name)

    def _join(self, mode, name, on=None, using=None):
        name = self.db.quote(name)
        name = sql.join(self, mode, name, on, using)
        return Relation(self.db, name, 'join', None)


class Relatiomat(object):

    def __init__(self, db):
        self.__db = db
        self._cache = {}

    def __get_relation(self, name):
        name = '.'.join(name.split('__', 1))
        try:
            return self._cache[name]
        except KeyError:
            rtype, pk = self.__db.get_relation_info(name)
            r = Relation(self.__db, name, rtype, pk or ())
            self._cache[name] = r
            return r

    __call__ = __get_relation
    __getattr__ = __get_relation
    __getitem__ = __get_relation


def logquery(meth):

    rx_strip_query = re.compile(r' *\n *')

    @wraps(meth)
    def wrapper(db, query, *args):
        stripped = rx_strip_query.sub(' ', query.strip())
        params, = args or ((),)
        log.debug('%s: %s %r' % (meth.__name__, stripped, params))
        return meth(db, query, *args)

    return wrapper


class Database(object):

    thread_local_config = False

    placeholder = None
    quote_char = '"'
    schema_separator_char = None
    script_func = 'execute'

    def __init__(self):
        self._tl = threading.local()

    def connect(self):
        raise NotImplementedError

    def disconnect(self):
        c = getattr(self._tl, 'connection', None)
        if c is not None:
            c.close()
        self._tl.connection = None

    def invalidate_relation_cache(self):
        """Invalidate the relation meta information cache.

        To be called when the database schema context or primary keys of tables
        have changed so that the cached meta information for relations may not
        be valid anymore.

        Note that this only affects the current thread's relation info cache.

        """
        self.relatiomat._cache.clear()

    @property
    def relatiomat(self):
        if getattr(self._tl, 'relatiomat', None) is None:
            self._tl.relatiomat = Relatiomat(self)
        return self._tl.relatiomat

    r = relatiomat
    t = relatiomat

    @property
    def connection(self):
        if getattr(self._tl, 'connection', None) is None:
            self.connect()
        return self._tl.connection

    def get_relation_info(self, name):
        raise NotImplementedError

    def begin(self):
        self.execute('BEGIN')

    def rollback(self):
        self.connection.rollback()

    def commit(self):
        self.connection.commit()

    @logquery
    def execute(self, query, params=()):
        c = self.connection.cursor()
        c.execute(query, params)
        return None if c.rowcount == -1 else c.rowcount

    @logquery
    def script(self, query):
        """Execute a parameterless query with multiple SQL statements

        This is only useful if a subclass' `execute` does not allow multiple
        statements in queries (e.g. `SQLiteDatabase`).

        """
        c = self.connection.cursor()
        f = getattr(c, self.script_func)
        f(query)

    @logquery
    def fetchall(self, query, params=()):
        c = self.connection.cursor()
        c.execute(query, params)
        return iter(c)

    @logquery
    def fetchone(self, query, params=()):
        c = self.connection.cursor()
        c.execute(query, params)
        return c.fetchone()

    @contextmanager
    def transaction(self):
        self.begin()
        try:
            yield
        except:
            self.rollback()
            raise
        self.commit()

    def quote(self, s):

        qc = self.quote_char
        ssc = self.schema_separator_char or ''
        parts = s.split(ssc) if ssc else [s]
        parts = ['%s%s%s' % (qc, part, qc) for part in parts]
        return ssc.join(parts)

    def insert_and_return_row(self, relation, row):
        raise NotImplementedError


class SQLiteDatabase(Database):

    placeholder = '?'
    script_func = 'executescript'

    def __init__(self, path):
        super(SQLiteDatabase, self).__init__()
        self._row_types = {}
        self._path = path

    def _row_factory(self, cursor, row):
        try:
            row_type = self._row_types[cursor.description]
        except KeyError:
            colnames = tuple(x[0] for x in cursor.description)
            row_type = namedtuple('DeenoSqliteRow', colnames)
            self._row_types[cursor.description] = row_type
        return row_type._make(row)

    def connect(self):
        self._tl.connection = sqlite3.connect(self._path)
        self._tl.connection.row_factory = self._row_factory
        self._tl.connection.isolation_level = None

    def get_relation_info(self, name):

        q = 'SELECT type FROM sqlite_master WHERE name = ?'
        p = [name]
        r = self.fetchone(q, p)
        if r is None or r.type not in ('table', 'view'):
            raise RelationNotFound('view or table %r not found' % name)

        pk = None
        if r.type == 'table':
            q = 'pragma table_info(%s)' % self.quote(name)
            rows = self.fetchall(q)
            pk = tuple(row.name for row in rows if row.pk)

        return r.type, pk

    def insert_and_return_row(self, relation, row):
        q, p = sql.insert(relation, row=row)
        c = self.connection.cursor()
        c.execute(q, p)
        rowid = c.lastrowid
        rows = relation.select(where={'rowid': rowid})
        return next(rows)


class PostgresDatabase(Database):

    placeholder = '%s'
    schema_separator_char = '.'

    def __init__(self, host, user, dbname):
        if isinstance(psycopg2, ImportError):
            raise psycopg2
        super(PostgresDatabase, self).__init__()
        self._host = host
        self._user = user
        self._dbname = dbname

    def connect(self):
        self._tl.connection = psycopg2.connect(host=self._host,
            user=self._user, database=self._dbname)
        self._tl.connection.cursor_factory = NamedTupleCursor
        self._tl.connection.autocommit = True

    def begin(self):
        self.connection.autocommit = False

    def rollback(self):
        self.connection.rollback()
        self.connection.autocommit = True

    def commit(self):
        self.connection.commit()
        self.connection.autocommit = True

    def get_relation_info(self, name):

        if '.' in name:
            schema, name = name.split('.', 1)
            schemas = [schema]
        else:
            q = "SHOW search_path"
            row = self.fetchone(q)
            schemas = row[0].replace(',', ' ').split()

        log.debug('look for %r in schemas %r' % (name, schemas))

        # Since the same base relation name may occur in multiple schemas, we
        # have to walk the search path to find the correct schema for a
        # relation.

        rtype, pk = None, None
        for schema in schemas:
            p = [schema, name]
            q = """
                SELECT 1 AS x FROM pg_catalog.pg_views
                WHERE schemaname = %s AND viewname = %s
            """
            row = self.fetchone(q, p)
            if row:
                rtype = 'view'
                break
            q = """
                SELECT 1 AS x FROM pg_catalog.pg_tables
                WHERE schemaname = %s AND tablename = %s
            """
            row = self.fetchone(q, p)
            if row:
                rtype = 'table'
                break

        if rtype is None:
            raise RelationNotFound('no view/table %r in current search path' %
                name)

        if rtype == 'table':
            q = """
                SELECT
                    a.attname
                FROM
                    pg_catalog.pg_attribute a,
                    pg_catalog.pg_class c,
                    pg_catalog.pg_index i
                WHERE
                    c.oid = %s::regclass AND
                    indrelid = c.oid AND
                    a.attrelid = c.oid AND
                    a.attnum = any(i.indkey)
                    AND indisprimary
            """
            p = [self.quote('%s.%s' % (schema, name))]
            rows = self.fetchall(q, p)
            pk = tuple(r[0] for r in rows)

        return rtype, pk

    def insert_and_return_row(self, relation, row):
        rows = relation.insert(row=row, returning='*')
        return next(rows)
